import logging
import pandas as pd
import datetime
import json
import operator
import csv
import os
import io

from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.core.urlresolvers import reverse, reverse_lazy
from django.core.serializers.json import DjangoJSONEncoder
from django.http import Http404, HttpResponseRedirect, HttpResponse
from django.shortcuts import render
from django.template import RequestContext
from django.utils.html import mark_safe
from django.views.generic.edit import CreateView, DeleteView
from django.conf import settings

import djcelery.models

from .models import Forecast
from .forms import ForecastForm
import apiv1 as fcst_api
from accountifie.query.query_manager import QueryManager
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory
import accountifie.toolkit.utils as utils
import accountifie.reporting.rptutils
from accountifie.common.table import get_table
from accountifie.toolkit.forms import LabelledFileForm, FileForm


logger = logging.getLogger('default')


@login_required
def forecast_index(request):
    context = {}
    context['title'] = 'Forecasts'
    context['content'] = get_table('forecasts')()
    return render(request, 'forecasts/forecast_list.html', context)


@login_required
def forecasts_reports(request):
    fcast_choices = [(f.id, f.label) for f in Forecast.objects.all()]
    context = {}
    context['fcasts'] = fcast_choices
    return render(request, 'forecasts/forecast_reports.html', context)


@login_required
def upload_gl(request):
    if request.method != "POST":
        forecast_obj = Forecast.objects.get(id=request.GET.get('forecast'))
        context = {'file_category': 'GL Projections'}
        context['file_type'] = '.csv'
        context['obj_label'] = forecast_obj.label
        form = FileForm()
        context['form'] = form

        return render(request, 'forecasts/upload_gl.html', context)
    else:
        form = FileForm(request.POST, request.FILES)
        if form.is_valid():
            upload = io.StringIO(unicode(request.FILES.values()[0].read()), newline=None)
            data = [row for row in csv.DictReader(upload)]
            forecast_obj = Forecast.objects.get(id=request.GET.get('forecast'))
            forecast_obj.hardcode_projections = data
            forecast_obj.save()
            return HttpResponseRedirect('/forecasts/forecast/%s' % forecast_obj.id)


def _order_projections(projs):
    cols = [x for x in projs[0].keys() if x not in ['Debit','Credit','Counterparty','Company']]

    def _idxer(label):
        lbls = label.split('M')
        yr = int(lbls[0])
        mth = int(lbls[1])
        return 2015 * yr + mth

    col_indexer = dict((x, _idxer(x)) for x in cols)
    return [x[0] for x in sorted(col_indexer.items(), key=operator.itemgetter(1))]


@login_required
def hardcode_projections(request):
    fcast_id = request.GET.get('fcast_id')
    hcode_projs = fcst_api.hardcode_projections(fcast_id, {})
    if len(hcode_projs) > 0:
        col_order = _order_projections(hcode_projs)
        context = {'cols': zip(['Debit', 'Credit', 'Counterparty', 'Company'] + col_order, ['nameFormatter']*4 + ['valueFormatter'] * len(col_order))}
        context['data_url'] = '/api/forecasts/hardcode_projections/%s?raw=true' % fcast_id
    else:
        context = {}
    return render(request, 'forecasts/bstrap_report.html', context)


@login_required
def all_projections(request):
    fcast_id = request.GET.get('fcast_id')
    projs = fcst_api.all_projections(fcast_id, {})
    if len(projs) > 0:
        col_order = _order_projections(projs)
        context = {'cols': zip(['Debit', 'Credit', 'Counterparty', 'Company'] + col_order, ['nameFormatter']*4 + ['valueFormatter'] * len(col_order))}
        context['data_url'] = '/api/forecasts/all_projections/%s?raw=true' % fcast_id
    else:
        context = {}

    return render(request, 'forecasts/bstrap_report.html', context)

def report_prep(request, id, version=None, strategy=None):

    as_of = request.GET.get('date', None)
    col_tag = request.GET.get('col_tag', None)

    format = request.GET.get('format', 'html')
    company_id = request.GET.get('company', utils.get_company(request))
    path = request.GET.get('path', None)
    report = accountifie.reporting.rptutils.get_report(id, company_id, version=version)

    if (company_id not in report.works_for):
        msg = "This ain't it. Report not available for %s" % report.company_name
        return render(request, '404.html', {'message': msg}), False

    report.configure(as_of=as_of, col_tag=col_tag, path=path)

    report.set_gl_strategy(strategy)
    return report, True, format


def create_report(request):

    fcast_id = request.GET.get('fcast_id')

    if 'date' in request.GET:
        dt = request.GET.get('date')
        rpt = request.GET.get('rpt')
        url = '/forecasts/reports/%s/%s?date=%s' % (fcast_id, rpt, dt)
    elif 'mth' in request.GET:
        mth = request.GET.get('mth')
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        url = '/forecasts/reports/%s/%s?col_tag=%sM%s' % (fcast_id, rpt, yr, mth)
    elif 'qtr' in request.GET:
        qtr = request.GET.get('qtr')
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        url = '/forecasts/reports/%s/%s?col_tag=%sQ%s' % (fcast_id, rpt, yr, qtr)    
    else:
        rpt = request.GET.get('rpt')
        yr = request.GET.get('yr')
        period = request.GET.get('period')
        url = '/forecasts/reports/%s/%s?col_tag=%s%s' % (fcast_id, rpt, yr, period)    

    clear_cache = request.GET.get('cl_cache', None)
    if clear_cache:
        if clear_cache == 'checked':
            url += '&clear_cache=True'

    return HttpResponseRedirect(url)


def fcast_report(request, fcast_id, rpt_id):

    fcast = Forecast.objects.get(id=fcast_id)

    strategy = QueryManagerStrategyFactory().get('forecast')
    strategy.set_cache(fcast_id=fcast_id, proj_gl_entries=fcast.get_gl_entries())

    report, is_report, format = report_prep(request, rpt_id, version='v1', strategy=strategy)
    if not is_report:
        return report

    report_data = report.calcs()
    
    if format == 'json':
        return HttpResponse(json.dumps(report_data, cls=DjangoJSONEncoder), content_type="application/json")
    elif format == 'csv':
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="report.csv"'

        writer = csv.writer(response)
        writer.writerow([''] + report.column_order)
        map_values = lambda x: '' if x == '' else str(x['text'])
        for row in report_data:
            writer.writerow([row['label']] + [map_values(row[col]) for col in [x for x in report.column_order if x in row]])
        return response
    elif format == 'html':
        # not bootstrap ... the more custom style for more complex reports
        context = report.html_report_context()
        context['query_string'] = request.META['QUERY_STRING']
        context['rows'] = []
        for rec in report_data:
            context['rows'] += report.get_row(rec)

        return render(request, 'report.html', context)
    else:
        msg = "Sorry. This format is not recognised : %s" % format
        return render(request, '404.html', {'message': msg}), False


@login_required
def forecast_detail(request, id):
    try:
        forecast = Forecast.objects.get(id=id)
    except Forecast.DoesNotExist:
        raise Http404

    form = ForecastForm(instance=forecast)

    context = dict(
        f=forecast,
        form=form
        )
    return render(request, 'forecasts/forecast_detail.html', context)






# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forecasts', '0005_auto_20160728_1628'),
    ]

    operations = [
        migrations.AlterField(
            model_name='projmodelv1param',
            name='proj_model',
            field=models.ForeignKey(to='forecasts.ProjectionModel'),
        ),
    ]



import json
from ast import literal_eval
from os.path import join as _join

from django.conf import settings
from django.template import Library

register = Library()



@register.filter
def file_path(args):
    _args = literal_eval(args)

    return _join(settings.DATA_URL, settings.PDFOUT_PATH,'forecast_%s_%s.csv' %(_args[0], _args[1]))



import logging
from urlparse import urlparse
import datetime
import pandas as pd
from decimal import Decimal

from django.db import models
from django.utils.safestring import mark_safe

import accountifie.toolkit.utils as utils
from accountifie.common.api import api_func
from jsonfield import JSONField

logger = logging.getLogger("default")


def _get_model_call(url):
    purl = urlparse(url)
    path = [x for x in purl.path.split('/') if x!='']
    if url == '' or path[0] != 'api':
        return 'BAD_PATH_URL', ''

    _qs = purl.query
    if _qs == '':
        qs = {}
    else:
        qs = dict((l[0], l[1]) for l in [el.split('=') \
                  for el in _qs.split('&')])
    return path, qs


def _get_calcs(url):
    path, qs = _get_model_call(url)
    if path == 'BAD_PATH_URL':
        return

    if len(path) == 3:
        return api_func(path[1], path[2], qstring=qs)
    elif len(path) == 4:
        return api_func(path[1], path[2], path[3], qstring=qs)


class Forecast(models.Model):
    label = models.CharField(max_length=50, blank=True)
    company = models.ForeignKey('gl.Company', default=utils.get_default_company)
    start_date = models.DateField()
    hardcode_projections = JSONField(blank=True)
    comment = models.TextField(blank=True)
    model = models.TextField(blank=True)

    class Meta:
        app_label = 'forecasts'
        db_table = 'forecasts_forecast'

    def __str__(self):
        return self.label

    @property
    def id_link(self):
        return mark_safe('<a href="/forecasts/forecast/%s/">%s' %( self.id, self.label))


    def get_projections(self):
        def _clean_field(k, v):
            if k not in ['Debit', 'Credit', 'Counterparty', 'Company']:
                return float(v)
            else:
                return v

        def _clean_proj(p):
            return dict((k, _clean_field(k, v)) for k, v in p.iteritems())

        hardcoded_projs = [_clean_proj(p) for p in self.hardcode_projections]
        calcd_projs = _get_calcs(self.model)
        if calcd_projs is None or calcd_projs == []:
            logger.info('Projections calc failed')

        projs = []

        if hardcoded_projs:
            projs += hardcoded_projs
        if calcd_projs:
            projs += calcd_projs
        return projs

    def get_gl_entries(self):
        projs = self.get_projections()
        if len(projs) == 0:
            return []

        proj_df = pd.DataFrame(projs)
        del proj_df['Company']
        trans_series = proj_df.groupby(['Debit', 'Credit', 'Counterparty']).sum().stack() 
        trans_df = pd.DataFrame({'amount': trans_series}).reset_index()
        trans_df['date'] = trans_df['level_3'].map(parse_date)
        trans_df['dateEnd'] = trans_df['date']
        trans_df['id'] = trans_df.index.map(lambda x: 'fcast_1_%s' %x)
        trans_df['comment'] = trans_df.apply(lambda row: '%s.%s.%s.%s' % (row['id'],row['Debit'],row['level_3'],row['Counterparty']),axis=1)

        trans_dict = trans_df.to_dict(orient='records')
        gl_entries = [get_gl_entry(tr) for tr in trans_dict]
        return gl_entries


# utility functions for creating GL entries for Forecast
def parse_date(label):
    lbls = label.split('M')
    yr = int(lbls[0])
    mth = int(lbls[1])
    return datetime.date(yr,mth,1).isoformat()

def get_lines(trans):
    debit = {'accountId': trans['Debit'], 'counterpartyId': trans['Counterparty'], 'amount': Decimal(trans['amount'])}
    credit = {'accountId': trans['Credit'], 'counterpartyId': trans['Counterparty'], 'amount': -Decimal(trans['amount'])}
    return [debit, credit]


def get_gl_entry(trans):
    gl_entry = dict((k, trans[k]) for k in ['id','date','dateEnd','comment'])
    gl_entry['lines'] = get_lines(trans)
    return gl_entry




from django.contrib import admin
from django.http import HttpResponseRedirect
from django.core.urlresolvers import reverse

from models import *


class ForecastAdmin(admin.ModelAdmin):
    list_display = ('id', 'label', 'start_date')

    def response_add(self, request, obj, post_url_continue=None):
        return HttpResponseRedirect(reverse('forecasts_index'))

admin.site.register(Forecast, ForecastAdmin)



"""
Strategy that delegates transaction querying to a snapshot of accountifie-svc.

It is closely modelled on the remote Strategy
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import time
import json
import urllib
import urllib2
from decimal import Decimal

from dateutil.parser import parse
from django.conf import settings
from query_manager_strategy import QueryManagerStrategy
from accountifie.common.api import api_func
import logging

DZERO = Decimal('0')
logger = logging.getLogger('default')


class QueryManagerSnapshotStrategy(QueryManagerStrategy):

    def set_cache(self, snapshot_time):
        self.snapshot_time = snapshot_time


    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.account_balances_for_dates(cmpny, account_ids, dates, with_counterparties, True, excl_contra, with_tags, excl_tags) for cmpny in company_list]
            return self.__merge_account_balances_for_dates_results(balances)

        client = accountifieSvcSnapshotClient(self.snapshot_time, company_id=company_id)
        date_indexed_account_balances = {}

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            interco_excluded_balances = client.balances(
                accounts=accounts_with_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_counterparties=interco_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_with_interco_exclusion) > 0 else []

            interco_included_balances = client.balances(
                accounts=accounts_without_interco_exclusion,
                from_date=start,
                to_date=end,
                with_counterparties=with_counterparties,
                excluding_contra_accounts=excl_contra,
                with_tags=with_tags,
                excluding_tags=excl_tags
            ) if len(accounts_without_interco_exclusion) > 0 else []

            balances = interco_excluded_balances + interco_included_balances
            date_indexed_account_balances[dt] = dict((balance['id'], float(balance['shift'])) for balance in balances)        
        return date_indexed_account_balances

    """
    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        interco_exempt_accounts = api_func('gl', 'externalaccounts')

        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            company_list = api_func('gl', 'company_list', company_id)
            balances = [self.transactions(cmpny, account_ids, from_date, to_date, chunk_frequency, with_counterparties, True, excl_contra) for cmpny in company_list]
            return self.__merge_transactions_results(balances)

        client = accountifieSvcSnapshotClient(company_id)

        interco_counterparties = self.get_interco_counterparties_for(company_id) if excl_interco else []
        accounts_with_interco_exclusion =    [id for id in account_ids if id not in interco_exempt_accounts] if excl_interco else []
        accounts_without_interco_exclusion = [id for id in account_ids if id in interco_exempt_accounts]     if excl_interco else account_ids

        interco_excluded_transactions = client.transactions(
            accounts=accounts_with_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            excluding_counterparties=interco_counterparties,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_with_interco_exclusion) > 0 else {}

        interco_included_transactions = client.transactions(
            accounts=accounts_without_interco_exclusion,
            from_date=from_date,
            to_date=to_date,
            chunk_frequency=chunk_frequency,
            with_counterparties=with_counterparties,
            excluding_contra_accounts=excl_contra
        ) if len(accounts_without_interco_exclusion) > 0 else {}

        # merge the two dictionaries
        transactions = interco_excluded_transactions.copy()
        transactions.update(interco_included_transactions)

        make_contras = lambda x: 'more than two' if len(x)>2 else '.'.join(x)

        formatted_transactions = []
        for account in transactions:
            for transaction in transactions[account]:
                if transaction['amount'] != '0.00':
                    formatted_transactions.append({
                        'date': parse(transaction['dateEnd'] if ('dateEnd' in transaction and transaction['dateEnd']) else transaction['date']).date(),
                        'id': transaction['id'],
                        'comment': transaction['comment'],
                        'account_id': account,
                        'contra_accts': make_contras(transaction['contraAccounts']),
                        'counterparty': transaction['counterparty'],
                        'amount': float(transaction['amount'])
                    })

        return formatted_transactions
    """

    def __merge_account_balances_for_dates_results(self, result_list):
        merged_results = {}

        for result in result_list:
            for date in result:
                merged_results[date] = merged_results[date] if date in merged_results else {}
                for account in result[date]:
                    existing_amount = merged_results[date][account] if account in merged_results[date] else float(0)
                    merged_results[date][account] = existing_amount + result[date][account]

        return merged_results

    """
    def __merge_transactions_results(self, result_list):
        return [transaction for result in result_list for transaction in result]

    def get_transaction(self, company_id, transaction_id):
        client = accountifieSvcSnapshotClient(company_id)
        return client.get_transaction(transaction_id)
    """
    


    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False

    @staticmethod
    def get_interco_counterparties_for(company):
        output = [cmpy['id'] for cmpy in api_func('gl', 'company') if cmpy['id']!=company and cmpy['cmpy_type']=='ALO']
        output += [x.lower() for x in output]
        return output

class accountifieSvcSnapshotClient(object):
    def __init__(self, snapshot_time, company_id='*'):
        self.company_id = company_id
        self.url_base = "%s/gl/%s/snapshot" % (self.__accountifie_svc_url(), company_id)
        self.snapshot_time = snapshot_time

    def __get(self, path, params=None):
        if params:
            params.update({'snapshotDate': self.snapshot_time})
        else:
            params = {'snapshotDate': self.snapshot_time}

        params = urllib.urlencode(params)
        url = '%s%s?%s' % (self.url_base, path, params)

        start_time = time.time()
        request = urllib2.Request(url)
        response = urllib2.urlopen(request)
        json_result = json.load(response)

        logger.info('Ran snapshots. %d seconds. URL=%s' % (time.time() - start_time, url))

        return json_result

    def balances(self, accounts, from_date=None, to_date=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None, with_tags=None, excluding_tags=None):
        from_date = None if from_date == '2000-01-01' else from_date
        account_balances = self.__get('/balances', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or []),
            'withTags': ','.join(with_tags or []),
            'excludingTags': ','.join(excluding_tags or []),
        })
        return account_balances

    """
    def transactions(self, accounts, from_date=None, to_date=None, chunk_frequency=None, with_counterparties=None, excluding_counterparties=None, excluding_contra_accounts=None):
        from_date = None if from_date == '2000-01-01' else from_date
        transaction_history = self.__get('/transactions', {
            'accounts': ','.join(accounts),
            'from': from_date,
            'to': to_date,
            'chunkFrequency': chunk_frequency,
            'withCounterparties': ','.join(with_counterparties or []),
            'excludingCounterparties': ','.join(excluding_counterparties or []),
            'excludingContraAccounts': ','.join(excluding_contra_accounts or [])
        })
        return transaction_history

    def get_transaction(self, transaction_id):
        return self.__get('/transaction/%s' % transaction_id)
    """
    
    @staticmethod
    def __accountifie_svc_url():
        return api_func('environment', 'variable', 'ACCOUNTIFIE_SVC_URL')



from betterforms.forms import Fieldset, BetterModelForm
from accountifie.toolkit.forms import JEditableMixin, BootstrapMixin, Html5Mixin

from .models import Forecast
from accountifie.toolkit.utils import random_color
from accountifie.gl.models import Company


class ForecastForm(BootstrapMixin, JEditableMixin, Html5Mixin, BetterModelForm):

    class Meta:
        model = Forecast
        fieldsets = (
            Fieldset('', (
                ('label',),
                ('start_date',),
                ('model',),
                ('comment',),
            )),
        )

    def __init__(self, *args, **kwargs):
        super(ForecastForm, self).__init__(*args, **kwargs)
        self.fields['label'].widget.attrs.update({'data-can_edit': 'false'})
        self.fields['comment'].widget.attrs.update({'data-can_edit': 'false'})



"""
Based on Andy Robinson's DoubleTalk accounting framework; used
with permission
"""

from decimal import Decimal
import types
import logging

import accountifie.gl.models
from django.contrib.contenttypes.models import ContentType
from django.conf import settings
import accountifie.query.query_manager_strategy_factory as QMSF


DZERO = Decimal('0')

logger = logging.getLogger('default')

class BusinessModelObject(object):
    """Something else which has an effect on the GL.
    
    BMOs will typically create General Ledger entries when
    saved.  They should efficiently update/change their
    GL entries and not create database churn unless there
    is a change.
    
    Make other models inherit from this as a mixin.

    Regrettably one then needs to override save() and
    delete() for the other models, until I find a cunning
    way to make it happen.
    
    """
    
    def delete_from_gl(self):
        "Find and remove all GL entries which depend on self"        
        bmo_id = '%s.%s' %(self.short_code, self.id)
        QMSF.getInstance().get().delete_bmo_transactions(self.company.id, bmo_id)


    def create_gl_transactions(self, trans):
        
        "Make any GL transactions this needs"
        for trandict in trans:
            d2 = trandict.copy()
            
            if not d2.has_key('date_end'):
                d2['date_end'] = d2['date']

            try:
                tags = ','.join(d2.pop('tags'))
            except:
                tags = ''

            d2['company'] = d2['company'].id if isinstance (d2['company'], accountifie.gl.models.Company) else d2['company']

            lines = d2.pop('lines')
            trans_id = d2.pop('trans_id')
            bmo_id = d2.pop('bmo_id')

            lines = [{
                        'account': account.id if isinstance (account, accountifie.gl.models.Account) else account,
                        'amount': "{0:.2f}".format(amount),
                        'counterparty': counterparty.id if isinstance (counterparty, accountifie.gl.models.Counterparty) else counterparty,
                        'tags': tags
                        } for account, amount, counterparty, tags in lines]
            
            QMSF.getInstance().get().create_gl_transactions(d2, lines, trans_id, bmo_id)

            """
            if GL_strategy == 'local':
                tran = accountifie.gl.models.Transaction(**d2)
                tran.source_object = self
                
                try:
                    if len(tran.comment) >= 100:
                        tran.comment = tran.comment[:99]
                except:
                    pass
                tran.save()    
            
                for (account, amount, counterparty, tags) in lines:
                    if type(account) in types.StringTypes:
                        account = accountifie.gl.models.Account.objects.get(id=account)
                    if type(counterparty) in types.StringTypes:
                        counterparty = accountifie.gl.models.Counterparty.objects.get(id=counterparty)
                    tran.tranline_set.create(account=account, amount=amount, counterparty=counterparty, tags=tags)
                
            elif GL_strategy == 'remote':
                # if we have a balanced transation then upsert
                if sum([line[1] for line in lines]) == DZERO:
                    QMSF.QueryManagerStrategyFactory().upsert_transaction({
                        'id': trans_id,
                        'bmo_id': '%s.%s' %(self.short_code, bmo_id),
                        #'object_id': tran.object_id,
                        'date': str(d2['date']),
                        'date_end': str(d2.get('date_end', None) or d2['date']),
                        'comment': d2['comment'],
                        'company': d2['company'].id if isinstance (d2['company'], accountifie.gl.models.Company) else d2['company'],
                        'type': None,
                        'lines': [{
                            'account': account.id if isinstance (account, accountifie.gl.models.Account) else account,
                            'amount': "{0:.2f}".format(amount),
                            'counterparty': counterparty.id if isinstance (counterparty, accountifie.gl.models.Counterparty) else counterparty,
                            'tags': tags
                        } for account, amount, counterparty, tags in lines]
                    })
            """    

    def update_gl(self):
        "Fix up GL after any kind of change"
        trans = self.get_gl_transactions()
        self.delete_from_gl()
        self.create_gl_transactions(trans)
        
    
    def get_gl_transactions(self):
        """Override this.  Create a list of transactions.
        
        Create a list of transactions in dictionary form
        which should exist.  It will create them all in the
        GL on saving if they do not already exist.
        """
        return []

    def get_company(self):
        return self.company.id
        
def on_bmo_save(sender, **kwargs):
    """Alternative method - maintain GL through a signal
    
    Master-detail BMOs can't use a simple save hook, and
    instead need to set a signal to call this'
    """
    sender.update_gl()





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('reporting', '0004_metricentry_label'),
        ('gl', '0002_transaction_bmo_id'),
        ('forecasts', '0003_remove_forecast_metrics'),
    ]

    operations = [
        migrations.CreateModel(
            name='ProjectionModel',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=30)),
                ('description', models.TextField()),
                ('as_of', models.DateTimeField()),
            ],
        ),
        migrations.CreateModel(
            name='ProjModelv1Param',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('frequency', models.CharField(max_length=20, choices=[(b'MONTHLY', b'Monthly'), (b'QUARTERLY', b'Quarterly'), (b'ANNUAL', b'Annual'), (b'ONE-OFF', b'One-off')])),
                ('window', models.CharField(max_length=20, choices=[(b'M1', b'Previous Month'), (b'M3', b'Rolling 3-month'), (b'M12', b'Rolling 12-month'), (b'A1', b'12 Month Lag')])),
                ('scaling', models.CharField(max_length=20, choices=[(b'PROP', b'Proportional')])),
                ('account', models.ForeignKey(to='gl.Account')),
                ('counterparty', models.ForeignKey(to='gl.Counterparty')),
                ('metric', models.ForeignKey(to='reporting.Metric')),
            ],
        ),
    ]



from django.conf.urls import url

import views
import jobs


urlpatterns = [
    url(r'forecasts/run', jobs.run_forecast, name='run_forecast'),

    url(r'forecasts/upload_gl/$', views.upload_gl, name='upload_gl'),
    url(r'forecasts/forecast/hardcode_projections', views.hardcode_projections, name="hardcode_projections"),
    url(r'forecasts/forecast/full_projections', views.all_projections, name="full_projections"),
    url(r'forecasts/reports/(?P<fcast_id>[()_a-zA-Z0-9]+)/(?P<rpt_id>[_a-zA-Z0-9]+)/$', views.fcast_report, name='fcast_report'),
    url(r'forecasts/forecast/(?P<id>[()_a-zA-Z0-9]+)/$', views.forecast_detail, name='forecasts_detail'),
    url(r'^forecasts/index/', views.forecast_index, name='forecasts_index'),
]



import csv
import json

from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import render
from django.template import RequestContext, Context
from django.core.serializers.json import DjangoJSONEncoder


import accountifie.reporting.rptutils as rptutils
import accountifie.toolkit.utils as utils

import logging

logger = logging.getLogger('default')



@login_required
def report(request, id):

    report, is_report = rptutils.report_prep(request, id)
    if not is_report:
            return report
        
    if report.from_file:
        try:
            report_data = json.loads(report.from_file)
        except:
            msg = "Sorry. file source is not recognised : %s" % report.from_file
            return render(request, 'rpt_doesnt_exist.html', {'message': msg})
    else:
        report_data = report.calcs()

    if report.format == 'json':
        return HttpResponse(json.dumps(report_data, cls=DjangoJSONEncoder), content_type="application/json")
    elif report.format == 'csv':
        response = HttpResponse(content_type='text/csv')
        response['Content-Disposition'] = 'attachment; filename="report.csv"'

        writer = csv.writer(response)
        writer.writerow([''] + report.column_order)
        map_values = lambda x: '' if x == '' else str(x['text']).replace('(', '-').replace(')', '')
        
        for row in report_data:
            if row['fmt_tag'] != 'header':
                writer.writerow([row['label']] + [map_values(row[col]) for col in report.column_order])
        return response
    elif report.format == 'html':
        # not bootstrap ... the more custom style for more complex reports
        context = report.html_report_context()
        context['query_string'] = request.META['QUERY_STRING']
        context['rows'] = []
        for rec in report_data:
            context['rows'] += report.get_row(rec)

        return render(request, 'report.html', context)
    else:
        msg = "Sorry. This format is not recognised : %s" % format
        return render(request, '404.html', {'message': msg}), False



import gl_helpers

import datetime
import json
import urllib2
from dateutil.parser import parse
from dateutil.relativedelta import relativedelta
from decimal import Decimal, getcontext, ROUND_HALF_UP, ROUND_HALF_DOWN
from bisect import insort
import locale
from string import letters
import operator
import re

import csv, os
from pprint import pprint

from django.db import models
from django.conf import settings

import datefuncs

locale.setlocale(locale.LC_ALL, '')



import logging
logger = logging.getLogger('default')


#use this frequently in rounding so give it a name
DZERO = Decimal("0.00")
HUNDREDTH = Decimal("0.01")

"""
def periods(period_tags):
    periods = {}
    for tag in period_tags:
        period_tag = period_tags[tag]
        periods[tag] = {'start': start_of_period(period_tag), 'end': end_of_period(period_tag)}
    return periods
"""

def day_before(d):
    return d - datetime.timedelta(days=1)

from pandas.tseries.offsets import BDay
def prev_busday(d):
    return (d - BDay(1)).date()


def entry(x, link='', ccy_fmt=''):
    return {'text': fmt(x, values_fmt=ccy_fmt), 'link': link}

def acct_history_link(id):
    from_dt = datefuncs.start_of_prev_year(datetime.datetime.now().year).isoformat()
    return ('/reporting/history/account/%s/?from=%s' % (id, from_dt))

def path_history_link(path):
    from_dt = datefuncs.start_of_prev_year(datetime.datetime.now().year).isoformat()
    return ('/reporting/history/path/%s/?from=%s' % (path.replace('.', '_'), from_dt))

def path_balances_link(path, date='today'):
    return ('/gl/path/%s/balances/?date=date' % path.replace('.','_'))

def fmt(stuff, values_fmt=None):
    "How do we want money formatted?"
    if stuff is None:
        return '----'
    elif isinstance(stuff, basestring):
        return stuff
    else:
        #accounting format, hacky, must be a recipe for this.
        #just seeing if brackets line up.
        if abs(stuff) < 0.1:
            txt = '-'
        elif stuff < 0:
            txt = '(' + locale.format("%0.0f", abs(stuff), grouping=True) + ')'
        else:
            txt = locale.format("%0.0f", stuff, grouping=True) + ''
        if values_fmt:
            return values_fmt + txt
        else:
            return txt

def unfmt(x):
    x = x.replace('$', '')
    if x == '-':
        return 0.0
    else:
        return float(x.replace(',','').replace(')','').replace('(','-'))





def get_dates(dt):
    if datefuncs.is_period_id(dt):
        start = datefuncs.start_of_period(dt)
        end = datefuncs.end_of_period(dt)
    elif type(dt) in [str, unicode] and dt[-4:]=='_YTD':
        end = parse(dt[:-4])
        start = datetime.date(end.year, 1, 1)
    elif type(dt) in [str, unicode] and dt[0] == 'D':
        d = parse(dt[1:]).date()
        start = prev_busday(d)+datetime.timedelta(days=1)
        end = d
    elif dt=='today':
        start = settings.DATE_EARLY
        end = datetime.datetime.now().date()
    else:
        start = settings.DATE_EARLY
        if type(dt)==str:
            end = parse(dt).date()
        else:
            end = dt
    return start, end

def get_dates_dict(dt):
    if type(dt)==dict:
        if 'start' in dt and 'end' in dt:
            #already in format
            return dt
            
    start, end = get_dates(dt)
    return { 'start': start, 'end': end }




def to_dict(dataset):
    "Make a dict with column 1 as key and cols 2,3,4 etc as value."
    d = {}
    for row in dataset:
        first, rest = row[0], row[1:]
        d[first] = rest
    return d

def denoneify(num):
    if num is None:
        return DZERO
    else:
        return num

def safe_sum(seq):
    "Safe with None"
    try:
        return sum(seq)
    except TypeError:
        seq2 = []
        for elem in seq:
            if elem is not None:
                seq2.append(elem)

        return sum(seq2)







def get_columns(request):
    if request.GET.has_key('columns'):
        return request.GET.get('columns').split('.')
    else:
        return None



def files_for_dir(datadir):
    return [os.path.join(datadir, name) for name in os.listdir(datadir) if os.path.isfile(os.path.join(datadir, name))]

def csv_to_modelattr(open_file, name_cleaner=None, company=gl_helpers.get_default_company()):
    '''takes the fields and values in the CSV and transforms them into a list of dicts where the keys
     will match model attributes. for example Full Description becomes full_description'''
    if name_cleaner == None:
        name_cleaner = lambda name: name
    f_csv = csv.DictReader(open_file) 
    csv_to_modelattr = dict([(name, name_cleaner(name)) for name in f_csv.fieldnames])
    csv_to_modelattr['company_id'] = company

    return [dict([(csv_to_modelattr[name], value) for name, value in row.items() if name in csv_to_modelattr]) for row in f_csv]
    
def get_foreignkeys(model):
    return dict(((f.name, f.rel.to) for f in model._meta.fields if f.__class__ == models.ForeignKey))

def get_fk_attr(model):
    return [f.name for f in model._meta.fields if f.__class__ == models.ForeignKey]

def get_pk_name(model):
    return model._meta.pk.name
    
def instance_nonrel_data(row, model, name_cleaner=None, value_cleaner=None):
    model_flds =  model._meta.get_all_field_names()
    instance_data_no_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in row.items() if name_cleaner(name)
                        and name_cleaner(name) not in get_fk_attr(model) and name_cleaner(name) in model_flds)
    return model(**instance_data_no_fk)
    
def set_foreignkeys(instance, row, model, name_cleaner=None, value_cleaner=None):
    if get_foreignkeys(model): 
        instance_fk = dict((name_cleaner(name), value_cleaner(name, value)) for name, value in row.items() if name_cleaner(name) 
                        and name_cleaner(name) in get_fk_attr(model))
        for fk in get_foreignkeys(model).items():
            if instance_fk.has_key(fk[0]):
                try:
                    related = fk[1].objects.get(pk=instance_fk[fk[0]])
                    setattr(instance, fk[0], related)
                except:
                    logger.error("No ForeignKey %s %s.  %s" % (fk[0], str(fk[1]), instance_fk))
    return instance
    
def dirty_key(row, model=None, unique=None, name_cleaner=None, value_cleaner=None):
    dirty = [name_cleaner(k) for k in row.keys() 
                if name_cleaner(k) not in [f.name for f in [field for field in model._meta.fields 
                        if field not in get_fk_attr(model)]] 
                                if name_cleaner(k)]

    return dirty
    
def create_instance(row, model, name_cleaner=None, value_cleaner=None, unique=None, exclude=[], company=gl_helpers.get_default_company()):
    row['company'] = company
    non_rel_instance = instance_nonrel_data(row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)
    if non_rel_instance.id in exclude:
        return None

    full_instance = set_foreignkeys(non_rel_instance, row, model, name_cleaner=name_cleaner, value_cleaner=value_cleaner)    

    return unique(full_instance)
    

def random_color(dark=True):
    import random
    r = lambda: random.randrange(10,130 if dark else 255, 10)
    return '#%02X%02X%02X' % (r(),r(),r())



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forecasts', '0009_auto_20160916_1116'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='projmodelv1param',
            name='account',
        ),
        migrations.RemoveField(
            model_name='projmodelv1param',
            name='contra',
        ),
        migrations.RemoveField(
            model_name='projmodelv1param',
            name='counterparty',
        ),
        migrations.RemoveField(
            model_name='projmodelv1param',
            name='metric',
        ),
        migrations.RemoveField(
            model_name='projmodelv1param',
            name='proj_model',
        ),
        migrations.DeleteModel(
            name='ProjectionModel',
        ),
        migrations.DeleteModel(
            name='ProjModelv1Param',
        ),
    ]



import os
import logging
import csv

from django.http import HttpResponseRedirect
from django.conf import settings

from models import Forecast
from accountifie.celery import background_task
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory
import accountifie.toolkit.utils as utils
import accountifie.reporting.rptutils as rptutils


logger = logging.getLogger('default')

REPORTPACK = ['IncBalanceSheet', 'IncomeStatement', 'Cashflow']

def run_forecast(request):
    # create csv output

    config = {}
    config['fcast_id'] = request.GET['forecast']
    config['col_tag'] = request.GET['col_tag']
    
    task_name = 'forecast_reportpack'

    task_id = background_task(task=task_name,
                              calc=_run_forecast,
                              config=config).id
    return HttpResponseRedirect('/tasks/list/')


def _run_forecast(*args, **kwargs):
    config = kwargs['config']
    col_tag = config['col_tag']
    company_id = config.get('company_id', utils.get_default_company())
    fcast_id = config['fcast_id']

    fcast = Forecast.objects.get(id=fcast_id)
    strategy = QueryManagerStrategyFactory().get('forecast')
    strategy.set_cache(fcast_id=fcast_id, proj_gl_entries=fcast.get_gl_entries())

    file_name = 'forecast_%s_reportpack.csv' %(fcast_id)
    path = os.path.join(settings.DATA_ROOT, settings.DOWNLOAD_PATH, file_name)
    
    if not os.path.isdir(settings.DOWNLOAD_OUT):
        os.makedirs(settings.DOWNLOAD_OUT)

    with open(path, 'wb') as f:
        writer = csv.writer(f)

        for report_id in REPORTPACK:
            writer.writerow([report_id])
            report = rptutils.get_report(report_id, company_id)
            report.configure(col_tag=col_tag)
            report.set_gl_strategy(strategy)
            report_data = report.calcs()

            writer.writerow([''] + report.column_order)
            map_values = lambda x: '' if x == '' else str(x['text'])
            for row in report_data:
                writer.writerow([row['label']] + [map_values(row[col]) for col in [x for x in report.column_order if x in row]])
            
            writer.writerow([])
            writer.writerow([])
            writer.writerow([])

    output_url = os.path.join(settings.DATA_URL, settings.DOWNLOAD_PATH, file_name)
    return {'download': output_url}



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0002_transaction_bmo_id'),
        ('forecasts', '0006_auto_20160728_1631'),
    ]

    operations = [
        migrations.AddField(
            model_name='projmodelv1param',
            name='contra',
            field=models.ForeignKey(related_name='param_contra', default='1001', to='gl.Account'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='projmodelv1param',
            name='account',
            field=models.ForeignKey(related_name='param_account', to='gl.Account'),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forecasts', '0004_projectionmodel_projmodelv1param'),
    ]

    operations = [
        migrations.AddField(
            model_name='projmodelv1param',
            name='label',
            field=models.CharField(default='', max_length=20),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='projmodelv1param',
            name='proj_model',
            field=models.ForeignKey(to='forecasts.ProjectionModel', null=True),
        ),
    ]



from multipledispatch import dispatch


from accountifie.forecasts.models import Forecast
from accountifie.forecasts.serializers import ForecastSerializer
import logging

logger = logging.getLogger('default')


@dispatch(dict)
def forecast(qstring):
    qs = Forecast.objects.all().order_by('-start_date')
    return list(ForecastSerializer(qs, many=True).data)


@dispatch(str, dict)
def forecast(label,qstring):
    qs = Forecast.objects.filter(label=label).first()
    return ForecastSerializer(qs).data


def hardcode_projections(fcast_id, qstring):
    return Forecast.objects.get(id=fcast_id).hardcode_projections


def all_projections(fcast_id, qstring):
    return Forecast.objects.get(id=fcast_id).get_projections()



"Re-saves all things which might produce GL transactions."
import inspect
import importlib
import logging
import datetime

from django.db import transaction
from django.http import HttpResponseRedirect

from accountifie.gl.models import Transaction
from accountifie.gl.bmo import BusinessModelObject
from accountifie.common.api import api_func
from accountifie.celery import background_task
from accountifie.query.query_manager_strategy_factory import QueryManagerStrategyFactory

logger = logging.getLogger('default')


def recalculate(request):
    task_name = 'ledger-recalculate'
    task_id = background_task(task=task_name, calc=_recalculate).id
    status_url = 'background_task/status/%s' % task_id
    return HttpResponseRedirect('/tasks/list/')


def _recalculate(*args, **kwargs):
    klasses = []
    kl_paths = api_func('environment', 'variable_list', 'BMO_MODULES')

    # find all the BMO classes
    for path in kl_paths:
        for name, kl in inspect.getmembers(importlib.import_module(path), inspect.isclass):
            if BusinessModelObject in kl.__bases__:
                klasses.append(kl)

    with transaction.atomic():
        Transaction.objects.all().delete()

        for cmpny in [c['id'] for c in api_func('gl', 'company')]:
            QueryManagerStrategyFactory().erase(cmpny)

        logger.info("Recalculate -- deleted all transactions")
        QueryManagerStrategyFactory().set_fast_inserts('*', True)
        for klass in klasses:
            logger.info('Recalculate -- working on %s' % klass)
            qs =klass.objects.all()
            for obj in qs:
                try:
                    obj.update_gl()
                except:
                    logger.error('Recalc failed on %s - %s' % (klass._meta.model_name, str(obj)))
            logger.info('Recalculate -- finished %s' % klass)
        QueryManagerStrategyFactory().set_fast_inserts('*', False)
        QueryManagerStrategyFactory().take_snapshot('*')

    logger.info("Recalculate -- updated %d transactions" % qs.count())
    return



"""
Strategy that performs transaction querying from in memory snapshots.

See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""
import itertools
from functools import partial

import dateutil.parser
import pandas as pd
import datetime

import accountifie.toolkit.utils as utils
from query_manager_strategy import QueryManagerStrategy
import query_manager
from accountifie.common.api import api_func
from accountifie.forecasts.models import Forecast

import logging

logger = logging.getLogger('default')


INCEPTION = datetime.date(2012,12,31)
FOREVER = datetime.date(2099,12,31)


def get_trans_data(tr):
    trans_keys = ['comment', 'dateEnd', 'date', 'type', 'id']
    return dict((k, v) for k, v in tr.iteritems() if k in trans_keys)

def get_contra(tr, account_id):
    return ','.join([x['accountId'] for x in tr['lines'] if x['accountId'] != account_id])


def parse_cache(data, company_id='INC'):
    lines = [[dict(l.items() + get_trans_data(tr).items() + [('contra_accts', get_contra(tr, l['accountId']))]) for l in tr['lines']] for tr in data]

    col_rename = {'accountId': 'account_id', 'dateEnd': 'date_end', 'counterpartyId': 'counterparty'}
    cache = pd.DataFrame(list(itertools.chain(*lines)))
    cache.rename(columns=col_rename, inplace=True)
    cache['date'] = cache['date'].map(lambda x: dateutil.parser.parse(x).date())
    if 'date_end' in cache.columns:
        cache['date_end'] = cache['date_end'].map(lambda x: dateutil.parser.parse(x).date())
        cache['date_end'].fillna(cache['date'], inplace=True)
    else:
        cache['date_end'] = cache['date']
    cache['amount'] = cache['amount'].astype(float)
    if 'Company' in cache.columns:
        cache['columns'] = cache['columns'].fillna(company_id)
    else:
        cache['company'] = company_id

    return cache


class QueryManagerForecastStrategy(QueryManagerStrategy):
    def set_cache(self, hist_strategy=None, fcast_id=None, proj_gl_entries=None):
        self.hist_strategy = hist_strategy
        self.forecast = Forecast.objects.get(id=fcast_id)
        self.cached_projections = self.forecast.get_projections()
        self.calc_balances()
        self.shifts = None
        self.balances = None

    def calc_balances(self):
        # limited task is to 
        # 1) create changes per period... columns = 2016M01, 2016M02 etc
        # 2) create cumulative changes as of each month end

        # changes per period come direct from the projections
        entries = pd.DataFrame(self.cached_projections)
        if entries.empty:
            self.balances = None
            return

        drop_cols = ['Counterparty', 'Company', 'Credit', 'Debit']
        credits = entries.copy()
        credits['account'] = credits['Credit']

        credits = credits[[c for c in credits.columns if c not in drop_cols]]
        debits = entries.copy()
        debits['account'] = debits['Debit']
        debits = debits[[c for c in debits.columns if c not in drop_cols]]

        for col in [c for c in credits.columns if c!='account']:
            credits[col] = credits[col] * -1

        self.shifts = pd.concat([credits, debits]).groupby('account').sum().fillna(0)
        balances_columns = dict((utils.end_of_period(col), col) for col in self.shifts.columns)
        sorted_months = sorted(balances_columns.keys())

        balances = {}
        for i in range(len(sorted_months)):
            label = sorted_months[i].isoformat()
            months_to_date = sorted_months[:i+1]
            shifts_cols = [balances_columns[l] for l in months_to_date]
            balances[label] = self.shifts[shifts_cols].sum(axis=1)

        self.balances = pd.DataFrame(balances)

        return None


    def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra, with_tags, excl_tags):
        self.calc_balances()
        date_indexed_account_balances = {}

        # cutoff should be last day of month eg 2016-3-31
        # get balances for the cutoff date
        hist_dates = dict((d, dates[d]) for d in dates if dates[d]['end'] <= self.forecast.start_date)
        proj_dates = dict((d, dates[d]) for d in dates if dates[d]['end'] > self.forecast.start_date)

        hist_qm = query_manager.QueryManager(gl_strategy=self.hist_strategy)
        hist_balances = hist_qm.pd_acct_balances(company_id, hist_dates, acct_list=account_ids, excl_contra=excl_contra, excl_interco=excl_interco, with_tags=with_tags, excl_tags=excl_tags)

        proj_start_dates = {'cutoff': {'start': INCEPTION, 'end': self.forecast.start_date }}
        proj_start_balances = hist_qm.pd_acct_balances(company_id, proj_start_dates, acct_list=account_ids, excl_contra=excl_contra, excl_interco=excl_interco, with_tags=with_tags, excl_tags=excl_tags)

        for dt in proj_dates:
            if utils.is_period_id(dt):
                if self.shifts is not None and dt in self.shifts.columns:
                    balances = self.shifts[dt].fillna(0).to_dict()
                else:
                    balances = {}
            else:
                if dt in self.balances.columns:
                    proj = {}
                    proj['cutoff'] = proj_start_balances['cutoff']
                    proj['fwd'] = self.balances[dt]
                    proj_df = pd.DataFrame(proj).fillna(0)
                    balances = proj_df.sum(axis=1).to_dict()
                else:
                    balances = {}
            date_indexed_account_balances[dt] = dict((balance, float(balances.get(balance,0))) for balance in balances if balance in account_ids)

        for dt in hist_dates:
            date_indexed_account_balances[dt] = hist_balances[dt].fillna(0).to_dict()

        return date_indexed_account_balances

    def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
        msg = "Transactions view not implemented for forecast strategies"
        return render(request, '404.html', {'message': msg}), False









    def proj_account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra):
        """

        """
        date_indexed_account_balances = {}
        entries = self.__pd_balances_prep(company_id, account_ids, excl_interco=excl_interco, excl_contra=excl_contra, with_counterparties=with_counterparties)
        if entries is None:
            return {dt: { account: 0.0 for account in account_ids } for dt in dates}

        for dt in dates:
            start = dates[dt]['start']
            end = dates[dt]['end']

            sub_entries = self.__depreciation_calcs(start, end, entries)

            col = sub_entries[['account_id', 'amount']].groupby('account_id').sum()['amount']
            date_indexed_account_balances[dt] = col.to_dict()

        return date_indexed_account_balances




    def get_gl_entries(self, company_id, account_ids, from_date=INCEPTION, to_date=FOREVER):
        return self.cached_projections


    def __pd_balances_prep(self, company_id, account_ids, excl_contra=None, excl_interco=False, with_counterparties=None):
        if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            excl_interco = True

        entries = self.get_gl_entries(company_id, account_ids)
        if entries is None or entries.empty:
            return None

        if excl_contra:
            for contra in excl_contra:
                entries = entries[entries.contra_accts.map(lambda x: contra not in x.split(','))]

        if excl_interco:
            entries = entries[~entries.apply(self.__inter_co, axis=1)]

        if with_counterparties:
          entries = entries[entries['counterparty'] in with_counterparties]

        if entries.empty:
            return None
        clean_end = lambda row: row['date_end'] if row['date_end'] else row['date']

        entries['date_end'] = entries.apply(clean_end, axis=1)


        return entries

    def __depreciation_calcs(self, start, end, entries):
        extend_comment = lambda row: row['comment'] + (' (prorated)' if row['deprec_factor'] < 1.0 else '')
        sub_entries = entries[(entries.date <= end) & (entries.date_end >= start)]
        _deprec_factor = partial(self.deprec_factor, start=start, end=end)

        if not sub_entries.empty:
            sub_entries.loc[:, 'deprec_factor'] = sub_entries.apply(_deprec_factor, axis=1)
            sub_entries.loc[:, 'comment'] = sub_entries.apply(extend_comment, axis=1)
            sub_entries.loc[:, 'amount'] *= sub_entries['deprec_factor']
            sub_entries.loc[:, 'amount'] = sub_entries['amount'].astype(float)
            #sub_entries.loc[:, 'date'] = sub_entries.apply(lambda row: min(row['date_end'], end), axis=1)
            sub_entries.loc[:, 'date_end'] = sub_entries.apply(lambda row: min(row['date_end'], end), axis=1)
        return sub_entries

    @staticmethod
    def monthly_chunk_periods(from_date, to_date):
        # dates should be from_date, to_date and the first and last days of every month included
        months = list(set(utils.monthrange(from_date, to_date)))
        dates = [(max(utils.start_of_month(x[1], x[0]), from_date), min(utils.end_of_month(x[1], x[0]), to_date)) for x in months]
        periods = [{'from': dt[0], 'to': dt[1]} for dt in dates]
        return periods

    @staticmethod
    def __inter_co(row):
        ext_accts = api_func('gl', 'externalaccounts')
        companies = [cmpy['id'] for cmpy in api_func('gl', 'company')]
        if row['account_id'] in ext_accts:
            return False
        if row['counterparty'] in companies:
            return True
        else:
            return False

    @staticmethod
    def deprec_factor(row, start, end):
        if row['date_end'] is None or row['date_end'] == row['date']:
            return 1.0
        days_outside = float(max((row['date_end'] - end).days, 0) + max((start - row['date']).days, 0))
        expense_period = float((row['date_end'] - row['date']).days + 1)
        deprec_factor = 1.0
        if expense_period > 0:
            deprec_factor -= days_outside / expense_period
        return deprec_factor



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('forecasts', '0008_auto_20160916_1108'),
    ]

    operations = [
        migrations.AlterField(
            model_name='forecast',
            name='model',
            field=models.TextField(blank=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import jsonfield.fields
import colorful.fields


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Forecast',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('label', models.CharField(max_length=50, blank=True)),
                ('start_date', models.DateField()),
                ('color_code', colorful.fields.RGBColorField(blank=True)),
                ('projections', jsonfield.fields.JSONField(default=dict, blank=True)),
                ('comment', models.TextField()),
            ],
            options={
                'db_table': 'forecasts_forecast',
            },
        ),
    ]




from .models import Forecast
from rest_framework import serializers


class ForecastSerializer(serializers.ModelSerializer):

    class Meta:
        model = Forecast
        fields = ('id_link', 'label', 'start_date', 'comment')



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import accountifie.toolkit.utils.gl_helpers


class Migration(migrations.Migration):

    dependencies = [
        ('gl', '0002_transaction_bmo_id'),
        ('forecasts', '0007_auto_20160729_1122'),
    ]

    operations = [
        migrations.RenameField(
            model_name='forecast',
            old_name='projections',
            new_name='hardcode_projections',
        ),
        migrations.RemoveField(
            model_name='forecast',
            name='color_code',
        ),
        migrations.AddField(
            model_name='forecast',
            name='company',
            field=models.ForeignKey(default=accountifie.toolkit.utils.gl_helpers.get_default_company, to='gl.Company'),
        ),
        migrations.AddField(
            model_name='forecast',
            name='model',
            field=models.URLField(blank=True),
        ),
        migrations.AlterField(
            model_name='forecast',
            name='comment',
            field=models.TextField(blank=True),
        ),
    ]



"""
Strategy that performs transaction querying locally.
The tried and tested way of doing things, though not as performant as
query_manager_remote_strategy.py.
See query_manager_strategy.py for interface docs.
Use query_manager_strategy_factory.py to get an instance of this class.
"""

import dateutil.parser
import pandas as pd
import accountifie.gl.cache
from accountifie.common.api import api_func
import accountifie.gl.models
import accountifie.toolkit.utils as utils

from datetime import datetime
from datetime import timedelta
from decimal import Decimal, ROUND_HALF_UP
from functools import partial
from query_manager_strategy import QueryManagerStrategy


class QueryManagerLocalStrategy(QueryManagerStrategy):

  def account_balances_for_dates(self, company_id, account_ids, dates, with_counterparties, excl_interco, excl_contra):
    date_indexed_account_balances = {}
    entries = self.__pd_balances_prep(company_id, account_ids, excl_interco=excl_interco, excl_contra=excl_contra, with_counterparties=with_counterparties)

    if entries is None:
      return {dt: { account: Decimal(0) for account in account_ids } for dt in dates}

    for dt in dates:
      start = dates[dt]['start']
      end = dates[dt]['end']

      sub_entries = self.__depreciation_calcs(start, end, entries)
      col = sub_entries[['account_id', 'amount']].groupby('account_id').sum()['amount']
      date_indexed_account_balances[dt] = col.to_dict()

    return date_indexed_account_balances

  def transactions(self, company_id, account_ids, from_date, to_date, chunk_frequency, with_counterparties, excl_interco, excl_contra):
    period_entries = []
    from_date = dateutil.parser.parse(from_date).date() if isinstance(from_date, basestring) else from_date
    to_date =   dateutil.parser.parse(to_date).date()   if isinstance(to_date, basestring) else to_date

    periods = [{'from': from_date, 'to': to_date}]
    if chunk_frequency == 'end-of-month':
      periods = self.monthly_chunk_periods(from_date, to_date)
    else:
      raise Exception('Unknown chunk frequency: %s' % chunk_frequency)

    cache = accountifie.gl.cache.get_cache(company_id)
    clean_end = lambda row: row['date_end'] if row['date_end'] else row['date']
    clean_id = lambda row: str(row['transaction_id'])
    for period in periods:
      start = period['from']
      end = period['to']

      entries = cache.get_gl_entries(account_ids, from_date=start, to_date=end)

      if entries is not None and not entries.empty:
        if with_counterparties:
          entries = entries[entries['counterparty'] in with_counterparties]

        if excl_contra and not entries.empty:
          for contra in excl_contra:
            entries = entries[entries.contra_accts.map(lambda x: contra not in x.split(','))]

        if excl_interco and not entries.empty:
          entries = entries[~entries.apply(self.__inter_co, axis=1)]

        if entries.empty:
          break

        entries['transaction_id'] = entries.apply(clean_id, axis=1)
        entries['date_end'] = entries.apply(clean_end, axis=1)
        sub_entries = self.__depreciation_calcs(start, end, entries)

        sub_entries.reset_index(drop=False, inplace=True)
        sub_entries.rename(columns={'transaction_id': 'id'}, inplace=True)
        period_entries.append(sub_entries)

    if len(period_entries) == 0:
      return None
    else:
      all_entries = pd.concat(period_entries, ignore_index=True)
      return all_entries.filter(items=['date', 'id', 'comment', 'account_id', 'contra_accts', 'counterparty', 'amount']).to_dict()

  def __pd_balances_prep(self, company_id, account_ids, excl_contra=None, excl_interco=False, with_counterparties=None):
    if api_func('gl', 'company', company_id)['cmpy_type'] == 'CON':
            excl_interco = True

    cache = accountifie.gl.cache.get_cache(company_id)

    entries = cache.get_gl_entries(account_ids)

    if entries is None or entries.empty:
      return None

    if excl_contra:
      for contra in excl_contra:
        entries = entries[entries.contra_accts.map(lambda x: contra not in x.split(','))]

    if excl_interco:
      entries = entries[~entries.apply(self.__inter_co, axis=1)]

    if with_counterparties:
      entries = entries[entries['counterparty'] in with_counterparties]

    if entries.empty:
      return None
    clean_end = lambda row: row['date_end'] if row['date_end'] else row['date']
    entries['date_end'] = entries.apply(clean_end, axis=1)

    return entries

  def create_gl_transactions(self, d2, lines, trans_id, bmo_id):
    tran = accountifie.gl.models.Transaction(**d2)
    tran.source_object = self
    
    try:
        if len(tran.comment) >= 100:
            tran.comment = tran.comment[:99]
    except:
        pass
    tran.save()    

    for (account, amount, counterparty, tags) in lines:
        if type(account) in types.StringTypes:
            account = accountifie.gl.models.Account.objects.get(id=account)
        if type(counterparty) in types.StringTypes:
            counterparty = accountifie.gl.models.Counterparty.objects.get(id=counterparty)
        tran.tranline_set.create(account=account, amount=amount, counterparty=counterparty, tags=tags)
                


  def delete_bmo_transactions(self, company_id, bmo_id):
    logger.info('Deleting transactions: bmp ID=%s, Company=%s' % (company_id, bmo_id))
    for t in accountifie.gl.models.Transaction.objects.filter(bmo_id=bmo_id):
      t.delete()

  def __depreciation_calcs(self, start, end, entries):
    extend_comment = lambda row: row['comment'] + (' (prorated)' if row['deprec_factor'] < Decimal('1') else '')
    sub_entries = entries[(entries.date <= end) & (entries.date_end >= start)]
    _deprec_factor = partial(self.deprec_factor, start=start, end=end)

    if not sub_entries.empty:
      sub_entries.loc[:, 'deprec_factor'] = sub_entries.apply(_deprec_factor, axis=1)
      sub_entries.loc[:, 'comment'] = sub_entries.apply(extend_comment, axis=1)
      sub_entries.loc[:, 'amount'] *= sub_entries['deprec_factor']
      sub_entries.loc[:, 'amount'] = sub_entries['amount'].map(lambda x: x.quantize(Decimal('0.01'), rounding=ROUND_HALF_UP))
      sub_entries.loc[:, 'date_end'] = sub_entries.apply(lambda row: min(row['date_end'], end), axis=1)
    return sub_entries

  @staticmethod
  def monthly_chunk_periods(from_date, to_date):
    # dates should be from_date, to_date and the first and last days of every month included
    months = list(set(utils.monthrange(from_date, to_date)))
    dates = [(max(utils.start_of_month(x[1], x[0]), from_date), min(utils.end_of_month(x[1], x[0]), to_date)) for x in months]
    periods = [{'from': dt[0], 'to': dt[1]} for dt in dates]
    return periods

  @staticmethod
  def __inter_co(row):
    ext_accts = api_func('gl', 'externalaccounts')
    companies = [cmpy['id'] for cmpy in api_func('gl', 'company')]
    if row['account_id'] in ext_accts:
      return False
    if row['counterparty'] in companies:
      return True
    else:
      return False


  @staticmethod
  def deprec_factor(row, start, end):
    if row['date_end'] is None or row['date_end'] == row['date']:
      return Decimal('1')
    days_outside = Decimal(max((row['date_end'] - end).days, 0) + max((start - row['date']).days, 0))
    expense_period = Decimal((row['date_end'] - row['date']).days + 1)
    deprec_factor = Decimal('1')
    if expense_period > 0:
      deprec_factor -= days_outside / expense_period
    return deprec_factor